@@ -1,4 +1,5 @@
 import asyncio
+import contextvars
 import inspect
 import logging
 from json import JSONDecodeError
@@ -32,6 +33,18 @@ import aiojobs
 logger = logging.getLogger(__name__)
 
 
+_jsonrpc_request_id = contextvars.ContextVar('_fastapi_jsonrpc__jsonrpc_request_id', default=None)
+_jsonrpc_method = contextvars.ContextVar('_fastapi_jsonrpc__jsonrpc_method', default=None)
+
+
+def get_jsonrpc_request_id() -> Any:
+    return _jsonrpc_request_id.get()
+
+
+def get_jsonrpc_method() -> Optional[str]:
+    return _jsonrpc_method.get()
+
+
 class Params(fastapi.params.Body):
     def __init__(
         self,
@@ -545,35 +558,56 @@ class MethodRoute(APIRoute):
     async def handle_http_request(self, http_request: Request):
         background_tasks = BackgroundTasks()
 
+        # noinspection PyTypeChecker
+        sub_response = Response(
+            content=None,
+            status_code=None,
+            headers=None,
+            media_type=None,
+            background=None,
+        )
+
         try:
             body = await self.parse_body(http_request)
         except Exception as exc:
             resp = await self.entrypoint.handle_exception_to_resp(exc)
+            response = self.response_class(content=resp, background=background_tasks)
         else:
             try:
-                resp = await self.handle_body(http_request, background_tasks, body)
+                resp = await self.handle_body(http_request, background_tasks, sub_response, body)
             except NoContent:
                 # no content for successful notifications
-                return Response(media_type='application/json', background=background_tasks)
+                response = Response(media_type='application/json', background=background_tasks)
+            else:
+                response = self.response_class(content=resp, background=background_tasks)
 
-        return self.response_class(content=resp, background=background_tasks)
+        response.headers.raw.extend(sub_response.headers.raw)
+        if sub_response.status_code:
+            response.status_code = sub_response.status_code
+
+        return response
 
     async def handle_body(
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         body: Any,
     ) -> dict:
         # Shared dependencies for all requests in one json-rpc batch request
         shared_dependencies_error = None
         try:
-            dependency_cache = await self.entrypoint.solve_shared_dependencies(http_request, background_tasks)
+            dependency_cache = await self.entrypoint.solve_shared_dependencies(
+                http_request,
+                background_tasks,
+                sub_response,
+            )
         except BaseError as error:
             shared_dependencies_error = error
             dependency_cache = None
 
         return await self.handle_req_to_resp(
-            http_request, background_tasks, body,
+            http_request, background_tasks, sub_response, body,
             dependency_cache=dependency_cache,
             shared_dependencies_error=shared_dependencies_error,
         )
@@ -582,12 +616,13 @@ class MethodRoute(APIRoute):
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         req: Any,
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
     ) -> dict:
         handler_coro = self.handle_req(
-            http_request, background_tasks, req,
+            http_request, background_tasks, sub_response, req,
             dependency_cache=dependency_cache,
             shared_dependencies_error=shared_dependencies_error,
         )
@@ -597,6 +632,7 @@ class MethodRoute(APIRoute):
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         req: Any,
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
@@ -623,24 +659,33 @@ class MethodRoute(APIRoute):
         # But if the methods have their own dependencies, they are resolved separately.
         dependency_cache = dependency_cache.copy()
 
-        values, errors, background_tasks, sub_response, _ = await solve_dependencies(
-            request=http_request,
-            dependant=self.func_dependant,
-            body=req['params'],
-            background_tasks=background_tasks,
-            dependency_overrides_provider=self.dependency_overrides_provider,
-            dependency_cache=dependency_cache,
-        )
+        jsonrpc_request_id_token = _jsonrpc_request_id.set(req.get('id'))
+        jsonrpc_method_token = _jsonrpc_method.set(req['method'])
 
-        if errors:
-            raise invalid_params_from_validation_error(RequestValidationError(errors))
+        try:
+            values, errors, background_tasks, _, _ = await solve_dependencies(
+                request=http_request,
+                dependant=self.func_dependant,
+                body=req['params'],
+                background_tasks=background_tasks,
+                response=sub_response,
+                dependency_overrides_provider=self.dependency_overrides_provider,
+                dependency_cache=dependency_cache,
+            )
 
-        result = await call_sync_async(self.func, **values)
+            if errors:
+                raise invalid_params_from_validation_error(RequestValidationError(errors))
+
+            result = await call_sync_async(self.func, **values)
+
+        finally:
+            _jsonrpc_request_id.reset(jsonrpc_request_id_token)
+            _jsonrpc_method.reset(jsonrpc_method_token)
 
         response = {
             'jsonrpc': '2.0',
             'result': result,
-            'id': req.get('id')
+            'id': req.get('id'),
         }
 
         # noinspection PyTypeChecker
@@ -731,7 +776,12 @@ class EntrypointRoute(APIRoute):
         self.entrypoint = entrypoint
         self.common_dependencies = common_dependencies
 
-    async def solve_shared_dependencies(self, http_request: Request, background_tasks: BackgroundTasks) -> dict:
+    async def solve_shared_dependencies(
+        self,
+        http_request: Request,
+        background_tasks: BackgroundTasks,
+        sub_response: Response,
+    ) -> dict:
         # Must not be empty, otherwise FastAPI re-creates it
         dependency_cache = {(lambda: None, ('', )): 1}
         if self.dependencies:
@@ -740,6 +790,7 @@ class EntrypointRoute(APIRoute):
                 dependant=self.shared_dependant,
                 body=None,
                 background_tasks=background_tasks,
+                response=sub_response,
                 dependency_overrides_provider=self.dependency_overrides_provider,
                 dependency_cache=dependency_cache,
             )
@@ -763,29 +814,50 @@ class EntrypointRoute(APIRoute):
     async def handle_http_request(self, http_request: Request):
         background_tasks = BackgroundTasks()
 
+        # noinspection PyTypeChecker
+        sub_response = Response(
+            content=None,
+            status_code=None,
+            headers=None,
+            media_type=None,
+            background=None,
+        )
+
         try:
             body = await self.parse_body(http_request)
         except Exception as exc:
             resp = await self.entrypoint.handle_exception_to_resp(exc)
+            response = self.response_class(content=resp, background=background_tasks)
         else:
             try:
-                resp = await self.handle_body(http_request, background_tasks, body)
+                resp = await self.handle_body(http_request, background_tasks, sub_response, body)
             except NoContent:
                 # no content for successful notifications
-                return Response(media_type='application/json', background=background_tasks)
+                response = Response(media_type='application/json', background=background_tasks)
+            else:
+                response = self.response_class(content=resp, background=background_tasks)
 
-        return self.response_class(content=resp, background=background_tasks)
+        response.headers.raw.extend(sub_response.headers.raw)
+        if sub_response.status_code:
+            response.status_code = sub_response.status_code
+
+        return response
 
     async def handle_body(
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         body: Any,
     ) -> dict:
         # Shared dependencies for all requests in one json-rpc batch request
         shared_dependencies_error = None
         try:
-            dependency_cache = await self.solve_shared_dependencies(http_request, background_tasks)
+            dependency_cache = await self.solve_shared_dependencies(
+                http_request,
+                background_tasks,
+                sub_response,
+            )
         except BaseError as error:
             shared_dependencies_error = error
             dependency_cache = None
@@ -803,7 +875,7 @@ class EntrypointRoute(APIRoute):
             for req in req_list:
                 job = await scheduler.spawn(
                     self.handle_req_to_resp(
-                        http_request, background_tasks, req,
+                        http_request, background_tasks, sub_response, req,
                         dependency_cache=dependency_cache,
                         shared_dependencies_error=shared_dependencies_error,
                     )
@@ -818,7 +890,7 @@ class EntrypointRoute(APIRoute):
         else:
             req = req_list[0]
             coro = self.handle_req_to_resp(
-                http_request, background_tasks, req,
+                http_request, background_tasks, sub_response, req,
                 dependency_cache=dependency_cache,
                 shared_dependencies_error=shared_dependencies_error,
             )
@@ -849,12 +921,13 @@ class EntrypointRoute(APIRoute):
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         req: Any,
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
     ) -> dict:
         handler_coro = self.handle_req(
-            http_request, background_tasks, req,
+            http_request, background_tasks, sub_response, req,
             dependency_cache=dependency_cache,
             shared_dependencies_error=shared_dependencies_error,
         )
@@ -864,6 +937,7 @@ class EntrypointRoute(APIRoute):
         self,
         http_request: Request,
         background_tasks: BackgroundTasks,
+        sub_response: Response,
         req: Any,
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
@@ -885,7 +959,7 @@ class EntrypointRoute(APIRoute):
             if match == Match.FULL:
                 # http_request is a transport layer and it is common for all JSON-RPC requests in a batch
                 return await route.handle_req(
-                    http_request, background_tasks, req,
+                    http_request, background_tasks, sub_response, req,
                     dependency_cache=dependency_cache,
                     shared_dependencies_error=shared_dependencies_error,
                 )
@@ -981,8 +1055,17 @@ class Entrypoint(APIRouter):
         for route in self.routes:
             route.dependency_overrides_provider = value
 
-    async def solve_shared_dependencies(self, http_request: Request, background_tasks: BackgroundTasks) -> dict:
-        return await self.entrypoint_route.solve_shared_dependencies(http_request, background_tasks)
+    async def solve_shared_dependencies(
+        self,
+        http_request: Request,
+        background_tasks: BackgroundTasks,
+        sub_response: Response,
+    ) -> dict:
+        return await self.entrypoint_route.solve_shared_dependencies(
+            http_request,
+            background_tasks,
+            sub_response,
+        )
 
     def add_method_route(
         self,
@@ -1005,7 +1088,7 @@ class Entrypoint(APIRouter):
         self,
         **kwargs,
     ) -> Callable:
-        def decorator(func: Callable) -> Callable:
+        def decorator(func: FunctionType) -> Callable:
             self.add_method_route(
                 func,
                 **kwargs,
